def validate_result(result):
    """
    Validate that the result matches the schema for `werkit.Manager`.
    """
    expected_keys = [
        "success",
        "result",
        "error",
        "error_origin",
        "duration_seconds",
        "runtime_info",
    ]
    if not isinstance(result, dict):
        raise ValueError(
            f"Expected Lambda result to be a dict with {','.join(expected_keys)}; got {result}"
        )
    result_keys = result.keys()
    missing_keys = [k for k in expected_keys if k not in result_keys]
    if len(missing_keys) > 0:
        raise ValueError(
            f"Lambda result contained keys {', '.join(result_keys)}; missing {','.join(missing_keys)}"
        )
    if not isinstance(result["success"], bool):
        raise ValueError(
            f"Expected result[\"success\"] to be a bool, got {result['success']}"
        )
    allowed_error_origins = ["compute", "system", "orchestration"]
    if result["error_origin"] not in [None] + allowed_error_origins:
        raise ValueError(
            f"Expected result[\"error_origin\"] to be {','.join(allowed_error_origins)}, or `None`"
        )


def wrap_result(serializable_result, start_time, duration_seconds, runtime_info=None):
    """
    Wrap the computation result in the `werkit.Manager` result schema.

    Args:
        serializable_result (object): Any serializable result.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The computation time.
        runtime_info (object): Serializable runtime metadata.
    """
    return {
        "success": True,
        "result": serializable_result,
        "error": None,
        "error_origin": None,
        "start_time": start_time.astimezone().isoformat(),
        "duration_seconds": duration_seconds,
        "runtime_info": runtime_info,
    }


def wrap_exception(
    exception, error_origin, start_time, duration_seconds=-1, runtime_info=None
):
    """
    Wrap an exception in the `werkit.Manager` result schema.

    Args:
        exception (Exception): The exception to wrap.
        error_origin (str): Either `"compute"`, `"system"` or
            `"orchestration"`. Refer to the `werkit.Manager` documentation for
            the definitions.
        start_time (datetime.datetime): The start time of the computation.
        duration_seconds (int): The time elapsed before the error occurred,
            if known.
        runtime_info (object): Serializable runtime metadata.
    """
    import traceback

    return {
        "success": False,
        "result": None,
        "error": traceback.format_exception(None, exception, exception.__traceback__),
        "error_origin": error_origin,
        "duration_seconds": duration_seconds,
        "start_time": start_time.astimezone().isoformat(),
        "runtime_info": runtime_info,
    }



import time
import os


def env_flag(env_var, default):
    import os

    environ_string = os.environ.get(env_var, "").strip().lower()
    if not environ_string:
        return default
    return environ_string in ["1", "true", "yes", "on"]


DELAY_SECONDS = int(os.environ.get("DELAY_SECONDS", "0"))
SHOULD_THROW = env_flag("SHOULD_THROW", False)


def wrap_result(serializable_result):
    """
    Simulate the effect of `werkit.Manager()`, which is not used here so this
    test service can be kept to one file.
    """
    return {
        "success": True,
        "result": serializable_result,
        "error": None,
        "error_origin": None,
        "duration_seconds": 1.0,
        "runtime_info": None,
    }


def handler(event, context):
    time.sleep(DELAY_SECONDS)
    if SHOULD_THROW:
        raise Exception("Whoops!")

    return wrap_result(event["input"] + sum(event["extra_args"]))



import time

# Import redis and rq lqzily to keep them soft dependencies.


class NotReady(Exception):
    pass


DEFAULT_QUEUE_NAME = "werkit-default"
JOB_ID_SEPARATOR = "____"


def clean(queue_name=DEFAULT_QUEUE_NAME, connection=None):
    from redis import Redis
    from rq import Queue
    from rq.registry import FinishedJobRegistry, FailedJobRegistry

    queue = Queue(queue_name, connection=connection or Redis())
    queue.empty()

    for registry in [FinishedJobRegistry(queue=queue), FailedJobRegistry(queue=queue)]:
        for job_id in registry.get_job_ids():
            registry.remove(job_id)


_clean_queue = clean


def invoke_for_each(
    fn,
    items,
    connection=None,
    queue_name=DEFAULT_QUEUE_NAME,
    clean=False,
    ensure_clean=True,
    as_kwarg=None,
    job_timeout="20m",
    result_ttl_seconds=60 * 60 * 24 * 30,
    args=(),
    kwargs={},
):
    """
    Invoke the given function on each of the given items.

    If `as_kwarg` is specified, the value is passed in as a keyword arg,
    otherwise as a positional arg.

    The additional `args` and `kwargs` are also provided to the function.

    When `ensure_empty` is true, raises an exception if there are already
    jobs in the queue.
    """
    from redis import Redis
    from rq import Queue

    if clean:
        _clean_queue(queue_name=queue_name, connection=connection)

    if ensure_clean:
        jobs = get_all_jobs(connection=connection, queue_name=queue_name)
        if len(jobs) > 0:
            raise ValueError(
                "Registry has not been cleaned! {} jobs were found".format(len(jobs))
            )

    queue = Queue(queue_name, connection=connection or Redis())
    for k, item in items.items():
        enqueue_kwargs = {
            "job_timeout": job_timeout,
            "result_ttl": result_ttl_seconds,
            "failure_ttl": result_ttl_seconds,
            "description": "{} {}".format(getattr(fn, "__name__", str(fn)), str(item)),
            "job_id": "{}{}{}".format(queue_name, JOB_ID_SEPARATOR, k),
        }
        if as_kwarg:
            out_fn_kwargs = kwargs.copy()
            out_fn_kwargs[as_kwarg] = item
            queue.enqueue(fn, args=args, kwargs=out_fn_kwargs, **enqueue_kwargs)
        else:
            queue.enqueue(fn, args=(item,) + args, kwargs=kwargs, **enqueue_kwargs)


def get_all_jobs(connection=None, queue_name=DEFAULT_QUEUE_NAME):
    from redis import Redis
    from rq import Queue
    from rq.job import Job
    from rq.registry import FinishedJobRegistry, FailedJobRegistry

    queue = Queue(queue_name, connection=connection or Redis())
    queued_jobs = queue.job_ids
    finished_jobs = FinishedJobRegistry(queue=queue).get_job_ids()
    failed_jobs = FailedJobRegistry(queue=queue).get_job_ids()
    return Job.fetch_many(
        queued_jobs + finished_jobs + failed_jobs, connection=connection
    )


def get_aggregate_status(
    connection=None, queue_name=DEFAULT_QUEUE_NAME, ret_jobs=False
):
    from rq.job import JobStatus

    jobs = get_all_jobs(connection=connection, queue_name=queue_name)
    if len(jobs) == 0:
        raise ValueError("No jobs found for queue {}".format(queue_name))
    statuses = set([job.get_status() for job in jobs])

    if set([JobStatus.FINISHED, JobStatus.FAILED]).issuperset(statuses):
        aggregate_status = JobStatus.FINISHED
    elif set([JobStatus.QUEUED]).issuperset(statuses):
        aggregate_status = JobStatus.QUEUED
    elif set(
        [JobStatus.QUEUED, JobStatus.STARTED, JobStatus.FINISHED, JobStatus.FAILED]
    ).issuperset(statuses):
        aggregate_status = JobStatus.STARTED
    else:
        aggregate_status = "other"

    return (aggregate_status, jobs) if ret_jobs else aggregate_status


def get_results(wait_until_done=False, queue_name=DEFAULT_QUEUE_NAME, connection=None):
    from rq.job import JobStatus

    while True:
        status, jobs = get_aggregate_status(connection=connection, ret_jobs=True)
        if status == JobStatus.FINISHED:
            return {j.id.split(JOB_ID_SEPARATOR)[-1]: j.result for j in jobs}
        elif wait_until_done:
            print("Status is {}; sleeping for {} seconds")
            time.sleep(30)
            continue
        else:
            raise NotReady()
